# Reproducing figures from the paper
# "Normalized mutual information is a biased measure for classification"
# (https://arxiv.org/abs/2307.01282)
import clustering_mi

# Figure 1

print("Figure 1:")

normalization_names = {"none": "I_0", "second": "NMI_0^A", "mean": "NMI_0^S"}

for normalization, name in normalization_names.items():
    NMI = clustering_mi.normalized_mutual_information(
        "data/2307.01282/fig_1_1.txt",
        normalization=normalization,
        variation="traditional",
    )  # Traditional mutual information
    print(f"{name}(c_1;g) = {NMI:.3f}")
    NMI = clustering_mi.normalized_mutual_information(
        "data/2307.01282/fig_1_2.txt",
        normalization=normalization,
        variation="traditional",
    )  # Traditional mutual information
    print(f"{name}(c_2;g) = {NMI:.3f}")

print()

# Figure 2

print("Figure 2:")

normalization_names = {"none": "I_0", "second": "NMI_0^A", "mean": "NMI_0^S"}

for normalization, name in normalization_names.items():
    NMI = clustering_mi.normalized_mutual_information(
        "data/2307.01282/fig_2_1.txt",
        normalization=normalization,
        variation="traditional",
    )  # Traditional mutual information
    print(f"{name}(c_1;g) = {NMI:.3f}")
    NMI = clustering_mi.normalized_mutual_information(
        "data/2307.01282/fig_2_2.txt",
        normalization=normalization,
        variation="traditional",
    )  # Traditional mutual information
    print(f"{name}(c_2;g) = {NMI:.3f}")

print()

# Figure 3

print("Figure 3:")

print("Contingency Table:")
print(clustering_mi._get_contingency_table("data/2307.01282/fig_3.txt"))

normalizations = ["second", "mean", "first"]
variations = ["traditional", "adjusted", "reduced"]

for normalization in normalizations:
    for variation in variations:
        NMI = clustering_mi.normalized_mutual_information(
            "data/2307.01282/fig_3.txt",
            normalization=normalization,
            variation=variation,
        )
        print(
            f"Normalization: {normalization}, Variation: {variation}, NMI = {NMI:.3f}"
        )



from __future__ import annotations

import numpy as np
import pytest

from clustering_mi._util import (
    _log_binom,
    _log_factorial,
    _log_Omega_EC,
    _minimize_golden_section_log,
)


def test__log_factorial():
    n = 5
    result = 1
    for i in range(1, n + 1, 1):
        result *= i
    assert _log_factorial(n) == pytest.approx(np.log(result))

    ns = [3, 4]
    results = [np.log(6), np.log(24)]
    for n, expected in zip(ns, results):
        assert _log_factorial(n) == pytest.approx(expected)


def test__log_binom():
    n = 5
    m = 2
    assert _log_binom(n, m) == pytest.approx(np.log(10))


def test__log_Omega_EC():
    rs = [1, 3, 4, 4]
    cs = [5, 2, 2, 3]
    result = 9.4314
    assert _log_Omega_EC(rs, cs) == pytest.approx(result)


def test__minimize_golden_section_log():
    f = lambda x: x**2 - 3 * x
    x_true = 1.5
    f_val_true = -2.25

    x, f_val = _minimize_golden_section_log(f, 0.01, 100)

    assert x == pytest.approx(x_true, rel=1e-2)
    assert f_val == pytest.approx(f_val_true, rel=1e-2)



from __future__ import annotations

import numpy as np
import pytest

from clustering_mi._input_output import _get_contingency_table
from clustering_mi.mutual_information import (
    _adjusted_mutual_information,
    _H_ng_G_alpha,
    _H_ngc_G_nc_alpha,
    _reduced_flat_mutual_information,
    _reduced_mutual_information,
    _stirling_mutual_information,
    _traditional_mutual_information,
    mutual_information,
    normalized_mutual_information,
)


def test__stirling_mutual_information():
    g = [1, 1, 2, 2, 2, 3, 3, 3]
    c = [1, 1, 1, 2, 2, 2, 2, 2]
    MI_true = 4.88058  # Computed using Mathematica in Tests.nb
    contingency_table = _get_contingency_table(c, g)
    MI = _stirling_mutual_information(contingency_table)
    assert pytest.approx(MI_true, rel=1e-5) == MI


def test__traditional_mutual_information():
    c = [1, 1, 1, 2, 2, 2, 2, 2]
    g = [1, 1, 2, 2, 2, 3, 3, 3]
    MI_true = 4.22239  # Computed using Mathematica in Tests.nb

    contingency_table = _get_contingency_table(c, g)
    MI = _traditional_mutual_information(contingency_table)
    assert pytest.approx(MI_true, rel=1e-5) == MI


def test__adjusted_mutual_information():
    c = [1, 1, 1, 2, 2, 2, 2, 2]
    g = [1, 1, 2, 2, 2, 3, 3, 3]
    MI_true = 2.74404  # Computed using Mathematica in Tests.nb

    contingency_table = _get_contingency_table(c, g)
    MI = _adjusted_mutual_information(contingency_table)
    assert pytest.approx(MI_true, rel=1e-5) == MI


def test__reduced_flat_mutual_information():
    c = [1, 1, 1, 2, 2, 2, 2, 2]
    g = [1, 1, 2, 2, 2, 3, 3, 3]
    MI_true = 1.07763  # Computed using Mathematica in Tests.nb
    contingency_table = _get_contingency_table(c, g)
    MI = _reduced_flat_mutual_information(contingency_table)
    assert pytest.approx(MI_true, rel=1e-5) == MI


def test__reduced_mutual_information():
    c = [1, 1, 1, 2, 2, 2, 2, 2]
    g = [1, 1, 2, 2, 2, 3, 3, 3]
    MI_true = 0.19313  # Computed using Mathematica in Tests.nb
    contingency_table = _get_contingency_table(c, g)
    MI = _reduced_mutual_information(contingency_table)
    assert pytest.approx(MI_true, rel=1e-5) == MI


def test__H_ng_G_alpha():
    ng = np.array([30, 20])
    alpha = 0.5
    H_ng_G_alpha_true = 4.35361  # Computed using Mathematica in Tests.nb
    H_ng_G_alpha = _H_ng_G_alpha(ng, alpha)
    assert H_ng_G_alpha == pytest.approx(H_ng_G_alpha_true, rel=1e-5)


def test__H_ngc_G_nc_alpha():
    ngc = np.array([[1, 0, 3], [2, 2, 0]])
    alpha = 0.5
    H_ngc_G_nc_alpha_true = 3.81796  # Computed using Mathematica in Tests.nb
    H_ngc_G_nc_alpha = _H_ngc_G_nc_alpha(ngc, alpha)
    assert H_ngc_G_nc_alpha == pytest.approx(H_ngc_G_nc_alpha_true, rel=1e-5)


def test__reduced_mutual_information():
    c = [1, 1, 1, 2, 2, 2, 2, 2]
    g = [1, 1, 2, 2, 2, 3, 3, 3]
    MI_true = 0.160127  # Computed using Mathematica in Tests.nb
    contingency_table = _get_contingency_table(c, g)
    MI = _reduced_mutual_information(contingency_table)
    assert pytest.approx(MI_true, rel=1e-5) == MI


def test_mutual_information():
    c = [1, 1, 1, 2, 2, 2, 2, 2]
    g = [1, 1, 2, 2, 2, 3, 3, 3]

    # - "reduced" (default): Reduced mutual information (RMI), reduction of https://arxiv.org/pdf/2405.05393, note that this can be (slightly) asymmetric.
    MI_true = 0.160127  # Computed using Mathematica in Tests.nb
    MI = mutual_information(c, g, variation="reduced")
    assert pytest.approx(MI_true, rel=1e-5) == MI

    # - "reduced_flat": Reduced mutual information (RMI), flat reduction of https://arxiv.org/pdf/1907.12581
    MI_true = 1.07763  # Computed using Mathematica in Tests.nb
    MI = mutual_information(c, g, variation="reduced_flat")
    assert pytest.approx(MI_true, rel=1e-5) == MI

    # - "adjusted": Adjusted mutual information (AMI), correcting for chance: https://jmlr.csail.mit.edu/papers/volume11/vinh10a/vinh10a.pdf
    MI_true = 2.74404  # Computed using Mathematica in Tests.nb
    MI = mutual_information(c, g, variation="adjusted")
    assert pytest.approx(MI_true, rel=1e-5) == MI

    # - "traditional": Traditional mutual information (MI), microcanonical
    MI_true = 4.22239  # Computed using Mathematica in Tests.nb
    MI = mutual_information(c, g, variation="traditional")
    assert pytest.approx(MI_true, rel=1e-5) == MI

    # - "stirling": Stirling's approximation of the traditional mutual information, equal to the mutual information of the corresponding probability distributions (times the number of objects).
    MI_true = 4.88058  # Computed using Mathematica in Tests.nb
    MI = mutual_information(c, g, variation="stirling")
    assert pytest.approx(MI_true, rel=1e-5) == MI

    # Bad variation types
    with pytest.raises(ValueError):
        mutual_information(c, g, variation="unknown")


def test_normalized_mutual_information():
    #  - "second" (default): Asymmetric normalization, measures how much the first labeling tells us about the second, as a fraction of all there is to know about the second labeling.
    c = [1, 1, 1, 2, 2, 2, 2, 2]
    g = [1, 1, 2, 2, 2, 3, 3, 3]
    NMI_true = 0.0202079
    NMI = normalized_mutual_information(c, g, normalization="second")
    assert pytest.approx(NMI_true, rel=1e-5) == NMI

    # - "first": Asymmetric normalization, measures how much the second labeling tells us about the first, as a fraction of all there is to know about the first labeling.
    NMI_true = 0.412743  # Computed using Mathematica in Tests.nb
    NMI = normalized_mutual_information(
        c, g, normalization="first", variation="adjusted"
    )
    assert pytest.approx(NMI_true, rel=1e-5) == NMI

    # - "mean": Symmetric normalization by the arithmetic mean of the two entropies.
    NMI_true = 0.48501  # Computed using Mathematica in Tests.nb
    NMI = normalized_mutual_information(
        c, g, normalization="mean", variation="stirling"
    )
    assert pytest.approx(NMI_true, rel=1e-5) == NMI

    # - "min": Normalize by the minimum of the two entropies.
    NMI_true = 0.727077  # Computed using Mathematica in Tests.nb
    NMI = normalized_mutual_information(
        c, g, normalization="min", variation="traditional"
    )
    assert pytest.approx(NMI_true, rel=1e-5) == NMI

    # - "max": Normalize by the maximum of the two entropies.
    NMI_true = 0.267954  # Computed using Mathematica in Tests.nb
    NMI = normalized_mutual_information(
        c, g, normalization="max", variation="reduced_flat"
    )
    assert pytest.approx(NMI_true, rel=1e-5) == NMI

    # - "geometric": Normalize by the geometric mean of the two entropies.
    NMI_true = 0.05534  # Computed using Mathematica in Tests.nb
    NMI = normalized_mutual_information(
        c, g, normalization="geometric", variation="reduced"
    )
    assert pytest.approx(NMI_true, rel=1e-5) == NMI

    # - "none": No normalization, returns the mutual information in bits.
    NMI_true = 0.160127  # Computed using Mathematica in Tests.nb
    NMI = normalized_mutual_information(c, g, normalization="none", variation="reduced")
    assert pytest.approx(NMI_true, rel=1e-5) == NMI

    # Bad normalization types
    with pytest.raises(ValueError):
        normalized_mutual_information(
            c, g, normalization="unknown", variation="reduced"
        )



# Basic example usage of the clustering_mi package as described in the README.md file.
import clustering_mi

# Load two labelings of the same set of objects

# As arrays:
labels1 = ["red", "red", "red", "blue", "blue", "blue", "green", "green"]
labels2 = [1, 1, 1, 1, 2, 2, 2, 2]

# As a contingency table, i.e., a matrix that counts label co-occurrences.
# Columns are the first labeling, rows are the second labeling:
contingency_table = [[3, 1, 0], [0, 2, 2]]

# Or as a space-separated file:
"""
red 1
red 1
red 1
blue 1
blue 2
blue 2
green 2
green 2
"""
filename = "data/example.txt"


# Defaults to the reduced mutual information (RMI)
mutual_information = clustering_mi.mutual_information(labels1, labels2)  # From lists
mutual_information = clustering_mi.mutual_information(contingency_table)  # From contingency table
mutual_information = clustering_mi.mutual_information(filename)  # Reads filename

print(f"Mutual Information: {mutual_information:.3f} (bits)")

# Compute other variants using the "variation" parameter.
# Correcting for chance (random permutations)
adjusted_mutual_information = clustering_mi.mutual_information(labels1, labels2, variation="adjusted")  
# Traditional mutual information
traditional_mutual_information = clustering_mi.mutual_information(labels1, labels2, variation="traditional")


# Symmetric normalization
normalized_mutual_information = clustering_mi.normalized_mutual_information(labels1, labels2, normalization="mean")
# "Normalized Mutual Information" most commonly refers to the Stirling-approximated mutual information
# divided by the mean of the entropies of the two labelings, although this is not our preferred measure.
normalized_stirling_mutual_information = clustering_mi.normalized_mutual_information(labels1, labels2, variation="stirling", normalization="mean")

print(f"(symmetric) Normalized Mutual Information (labels1 <-> labels2): {normalized_mutual_information:.3f}")

# Asymmetric normalization measures how much the first labeling tells us about the second,
# as a fraction of all there is to know about the second labeling.
# This form is appropriate when the second labeling is a "ground truth" and the first is a prediction.
asymmetric_normalized_mutual_information_1_2 = clustering_mi.normalized_mutual_information(labels1, labels2, normalization="second")
# Or when the first labeling is the ground truth and the second is a prediction.
asymmetric_normalized_mutual_information_2_1 = clustering_mi.normalized_mutual_information(labels1, labels2, normalization="first")

print(f"(asymmetric) Normalized Mutual Information (labels1 -> labels2): {asymmetric_normalized_mutual_information_1_2:.3f}")
print(f"(asymmetric) Normalized Mutual Information (labels2 -> labels1): {asymmetric_normalized_mutual_information_2_1:.3f}")


# Small helper functions
from __future__ import annotations

from collections.abc import Callable
from math import lgamma

import numpy as np
from numpy.typing import ArrayLike


def _log_factorial(n: ArrayLike | float) -> ArrayLike | float:
    """
    Compute the natural logarithm of the factorial of n.

    Parameters
    ----------
    n : ArrayLike | float
        The value for which to compute the log-factorial.

    Returns
    -------
    ArrayLike | float
        log(n!) (base e)
    """
    if isinstance(n, (list, np.ndarray)):
        return np.array([lgamma(x + 1) for x in n])
    return lgamma(n + 1)


def _log_binom(n: float, m: float) -> float:
    """
    Compute the natural logarithm of the binomial coefficient binomial(n, m).

    Parameters
    ----------
    n : float
        The number of items.
    m : float
        The number of items to choose.

    Returns
    -------
    float
        log(binomial(n, m)) (base e)
    """
    return _log_factorial(n) - _log_factorial(m) - _log_factorial(n - m)


def _log_Omega_EC(
    rs: ArrayLike,
    cs: ArrayLike,
    useShortDimension: bool = False,
    symmetrize: bool = False,
) -> float:
    """
    Approximate the log of the number of contingency tables with given row and column sums
    using the EC estimate of Jerdee, Kirkley, Newman (2022) https://arxiv.org/abs/2209.14869.

    Parameters
    ----------
    rs : list or int
        Row sums.
    cs : list or int
        Column sums.
    useShortDimension : bool, optional
        Whether to optimize the encoding by possibly swapping the definitions of rows and columns (default: True).
    symmetrize : bool, optional
        Whether to symmetrize the estimate (default: False).

    Returns
    -------
    float
        Estimate of log_Omega (base 2).
    """
    rs = np.array(rs)
    cs = np.array(cs)
    # Remove any zeros
    rs = rs[rs > 0]
    cs = cs[cs > 0]
    if len(rs) == 0 or len(cs) == 0:
        return -np.inf  # There are no tables
    if (
        useShortDimension
    ):  # Performance of the EC estimate is generally improved when there are
        # more rows than columns. If this is not the case, swap definitions around
        if len(rs) >= len(cs):
            return _log_Omega_EC(rs, cs, useShortDimension=False)
        return _log_Omega_EC(cs, rs, useShortDimension=False)
    if symmetrize:
        return (
            _log_Omega_EC(rs, cs, symmetrize=False)
            + _log_Omega_EC(cs, rs, symmetrize=False)
        ) / 2
    m = len(rs)
    N: float = np.sum(rs)
    if (
        len(cs) == N
    ):  # In this case, we may simply return the exact result (equivalent to alpha = inf)
        return _log_factorial(N + 1) - np.sum(_log_factorial(rs + 1))
    alphaC = (N**2 - N + (N**2 - np.sum(cs**2)) / m) / (np.sum(cs**2) - N)
    result = -_log_binom(N + m * alphaC - 1, m * alphaC - 1)
    for r in rs:
        result += _log_binom(r + alphaC - 1, alphaC - 1)
    for c in cs:
        result += _log_binom(c + m - 1, m - 1)
    return float(result / np.log(2))  # Convert to base 2


# Only including this due to issues with scipy.optimize.minimize_scalar
def _minimize_golden_section_log(
    f: Callable[[float], float], min_val: float, max_val: float, tol: float = 1e-5
) -> tuple[float, float]:
    """
    Optimize a function using the golden section search method. Optimization occurs in the logarithmic domain.

    Parameters
    ----------
    f : callable
        The function to optimize.
    min_val : float
        The lower bound of the search interval. Must be greater than 0.
    max_val : float
        The upper bound of the search interval. Must be greater than 0.
    tol : float, optional
        The tolerance for convergence in the log(parameter) (default: 1e-5).

    Returns
    -------
    tuple[float, float]
        The minimizing value of the parameter and the function value at that point.
    """
    a = np.log(min_val)
    b = np.log(max_val)

    phi = (1 + np.sqrt(5)) / 2  # Golden ratio
    resphi = 2 - phi

    c = a + resphi * (b - a)
    d = b - resphi * (b - a)

    while abs(c - d) > tol:
        if f(np.exp(c)) < f(np.exp(d)):
            b = d
            d = c
            c = a + resphi * (b - a)
        else:
            a = c
            c = d
            d = b - resphi * (b - a)

    return np.exp((a + b) / 2), f(np.exp((a + b) / 2))



# Checking input validity and converting each of the input types into a contingency table
# to be passed to the functions that calculate the mutual information.
from __future__ import annotations

import logging
from pathlib import Path

import numpy as np
from numpy.typing import ArrayLike

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def _get_contingency_table(
    input_data_1: ArrayLike | str, input_data_2: ArrayLike | None = None
) -> ArrayLike:
    """
    Get the contingency table T between two labelings from a pair of lists or the name of a space separated file of labels.
    T[r][s] counts the number of objects with label s in the first labeling and label r in the second labeling.

    Raises AssertionError for invalid inputs.

    Parameters
    ----------
    input_data_1 : ArrayLike or str
        First argument. This will either be a 2D array-like which specifies the contingency table,
        or a string which is the path to a file containing a list of pairs of labels,
        or a 1-D array-like of labels.
    input_data_2 : ArrayLike, optional
        Second argument. This can only be a 1-D array-like of labels in the case where the first argument is also such a list.

    Returns
    -------
    ArrayLike
        Contingency table as a 2D NumPy array.
    """

    # Print the types of the inputs
    if isinstance(input_data_1, str):
        # If the first argument is a string, it must be a file path
        # Check if the file exists
        if not Path(input_data_1).is_file():
            raise FileNotFoundError(f"File {input_data_1} does not exist.")
        # Read the file and convert it to a contingency table
        with Path(input_data_1).open(encoding="utf-8") as f:
            lines = f.readlines()
        labels = [line.strip().split() for line in lines]
        # Make sure there are two labels per line, report the offending line
        if not all(len(label) == 2 for label in labels):
            raise AssertionError(
                "Each line in the file must contain exactly two labels."
            )
        labels1, labels2 = zip(*labels)
        return _get_contingency_table(labels1, labels2)

    if isinstance(input_data_1, (list, np.ndarray, tuple)):
        # If the array is 2D, it is a contingency table
        input_data_1 = np.array(input_data_1)
        if input_data_1.ndim == 2:
            # Remove any empty rows or columns
            input_data_1 = input_data_1[~np.all(input_data_1 == 0, axis=1)]
            input_data_1 = input_data_1[:, ~np.all(input_data_1 == 0, axis=0)]
            if input_data_1.size == 0:
                raise AssertionError(
                    "The contingency table is empty after removing empty rows and columns."
                )
            if not np.issubdtype(input_data_1.dtype, np.integer):
                raise AssertionError(
                    "The contingency table must contain integer values."
                )
            if np.any(input_data_1 < 0):
                raise AssertionError(
                    "The contingency table must not contain negative values."
                )
            return input_data_1
        if input_data_1.ndim == 1:
            # If the array is 1D, it should be a list of labels, and we need a second argument
            if isinstance(input_data_2, (list, np.ndarray, tuple)):
                input_data_2 = np.array(input_data_2)
                if input_data_2.ndim != 1:
                    raise AssertionError(
                        "The second argument must be a 1D array-like of labels."
                    )
                # Create a contingency table from the two labelings
                labels1, labels2 = input_data_1, input_data_2
                unique_labels1 = np.unique(labels1)
                unique_labels2 = np.unique(labels2)
                # contingency_table = np.zeros((len(unique_labels1), len(unique_labels2)), dtype=int)
                contingency_table = np.zeros(
                    (len(unique_labels2), len(unique_labels1)), dtype=int
                )  # We flip this around in accordance with how the table is defined in https://arxiv.org/abs/2307.01282
                for i, label1 in enumerate(unique_labels1):
                    for j, label2 in enumerate(unique_labels2):
                        # contingency_table[i, j] = np.sum((labels1 == label1) & (labels2 == label2))
                        contingency_table[j, i] = np.sum(
                            (labels1 == label1) & (labels2 == label2)
                        )  # We flip this around in accordance with how the table is defined in https://arxiv.org/abs/2307.01282
                return contingency_table
            raise AssertionError(
                "If the first argument is a 1D array-like of labels, a second list of labels must be provided."
            )
        raise AssertionError("First argument has too many dimensions.")

    # If we reach this point, the input is invalid
    raise TypeError(
        "First argument must be a 2D array-like (contingency table), a 1D array-like of labels, or a string file path."
    )



from __future__ import annotations

import os

import numpy as np
import pytest

from clustering_mi._input_output import _get_contingency_table

abs_path = os.path.dirname(os.path.realpath(__file__))


def test__get_contingency_table():
    # This is a bit finicky, since the permutation of the group labels is arbitrary
    contingency_table_true = [[1, 0, 3], [2, 2, 0]]

    # Passing lists of labels
    labels1 = ["red", "red", "red", "blue", "blue", "blue", "green", "green"]
    labels2 = [1, 1, 1, 1, 2, 2, 2, 2]
    contingency_table_found = _get_contingency_table(labels1, labels2)
    assert np.array_equal(contingency_table_found, contingency_table_true)

    # Passing a file
    contingency_table_found = _get_contingency_table(abs_path + "/data/example.txt")
    assert np.array_equal(contingency_table_found, contingency_table_true)
    contingency_table_found = _get_contingency_table(
        abs_path + "/data/example_commas.txt"
    )
    assert np.array_equal(contingency_table_found, contingency_table_true)
    contingency_table_found = _get_contingency_table(
        abs_path + "/data/example_tabs.txt"
    )
    assert np.array_equal(contingency_table_found, contingency_table_true)

    with pytest.raises(AssertionError):
        _get_contingency_table(abs_path + "/data/example_missing_values.txt")

    # Passing a contingency table
    contingency_table_found = _get_contingency_table(contingency_table_true)
    assert np.array_equal(contingency_table_found, contingency_table_true)

    contingency_table_non_integer = [[3.2, 1, 0], [0, 2, 2]]
    with pytest.raises(AssertionError):
        _get_contingency_table(contingency_table_non_integer)

    contingency_table_negative = [[3, -1, 0], [0, 2, 2]]
    with pytest.raises(AssertionError):
        _get_contingency_table(contingency_table_negative)



# Compute the mutual information and associated entropies
from __future__ import annotations

import logging

import numpy as np
from numpy.typing import ArrayLike

from clustering_mi._input_output import _get_contingency_table

# from scipy.optimize import minimize_scalar # Used to optimize the alpha parameter in the Dirichlet-multinomial reduction
from clustering_mi._util import (
    _log_binom,
    _log_factorial,
    _log_Omega_EC,
    _minimize_golden_section_log,
)

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def _stirling_mutual_information(contingency_table: ArrayLike) -> float:
    """
    Compute the Stirling approximated mutual information for the given contingency table.
    This is the mutual information of the corresponding probability distributions (times the number of objects).

    Parameters
    ----------
    contingency_table : ArrayLike
        Contingency table T as a 2D NumPy array where T[r][s] counts the number of objects with label r in the ground truth g and label s in the candidate c.

    Returns
    -------
    float
        Mutual information (base 2)
    """

    # Compute summary information
    n: float = np.sum(contingency_table)
    nc = np.sum(contingency_table, axis=0)
    ng = np.sum(contingency_table, axis=1)

    MI = 0
    for r, ng_r in enumerate(ng):
        for s, nc_s in enumerate(nc):
            if contingency_table[r, s] > 0:
                MI += contingency_table[r, s] * np.log(
                    n * contingency_table[r, s] / (ng_r * nc_s)
                )

    # Convert to bits (log base 2)
    MI /= np.log(2)

    return float(MI)


def _traditional_mutual_information(contingency_table: ArrayLike) -> float:
    """
    Compute the unreduced microcanonical mutual information and entropies of the ground truth and candidate labelings.

    Parameters
    ----------
    contingency_table : ArrayLike
        Contingency table T as a 2D NumPy array where T[r][s] counts the number of objects with label r in the ground truth g and label s in the candidate c.

    Returns
    -------
    float
        Mutual information (base 2).
    """

    # Compute summary information
    n: float = np.sum(contingency_table)
    nc = np.sum(contingency_table, axis=0)
    ng = np.sum(contingency_table, axis=1)

    MI = (
        _log_factorial(n)
        - np.sum(_log_factorial(ng))
        - np.sum(_log_factorial(nc))
        + np.sum(_log_factorial(contingency_table.flatten()))
    )

    # Convert to bits (log base 2)
    MI /= np.log(2)

    return float(MI)


def _adjusted_mutual_information(contingency_table: ArrayLike) -> float:
    """
    Compute the adjusted mutual information, which corrects the mutual information for chance (random permutations of the labels).

    Parameters
    ----------
    contingency_table : ArrayLike
        Contingency table T as a 2D NumPy array where T[r][s] counts the number of objects with label r in the ground truth g and label s in the candidate c.

    Returns
    -------
    float
        Adjusted mutual information (base 2).
    """

    n: float = np.sum(contingency_table)
    nc = np.sum(contingency_table, axis=0)
    ng = np.sum(contingency_table, axis=1)
    qg = len(ng)
    qc = len(nc)

    EMI = 0
    for r in range(qg):
        for s in range(qc):
            for ngc in range(max(1, ng[r] + nc[s] - n), min(ng[r], nc[s]) + 1):
                EMI += (
                    ngc
                    * (np.log(n) + np.log(ngc) - np.log(ng[r]) - np.log(nc[s]))
                    * np.exp(
                        _log_binom(nc[s], ngc)
                        + _log_binom(n - nc[s], ng[r] - ngc)
                        - _log_binom(n, ng[r])
                    )
                )

    # Convert to bits (log base 2)
    EMI /= np.log(2)

    return float(_stirling_mutual_information(contingency_table) - EMI)


def _reduced_flat_mutual_information(contingency_table: ArrayLike) -> float:
    """
    Compute the reduced mutual information, using the flat reduction method of https://arxiv.org/pdf/1907.12581.

    Parameters
    ----------
    contingency_table : ArrayLike
        Contingency table T as a 2D NumPy array where T[r][s] counts the number of objects with label r in the ground truth g and label s in the candidate c.

    Returns
    -------
    float
        Reduced mutual information (base 2).
    """

    nc = np.sum(contingency_table, axis=0)
    ng = np.sum(contingency_table, axis=1)

    logOmega = _log_Omega_EC(nc, ng)

    return float(_traditional_mutual_information(contingency_table) - logOmega)


def _H_ng_G_alpha(ng: ArrayLike, alpha: float) -> float:
    """
    Compute the entropy of a vector of group sizes given the concentration parameter alpha.

    Parameters
    ----------
    ng : list or np.ndarray
        Vector of group sizes.
    alpha : float
        Concentration parameter.

    Returns
    -------
    float
        Entropy of the group sizes. (base e).

    """
    n: float = np.sum(ng)
    q = len(ng)

    H_ng = _log_binom(
        n + q * alpha - 1, q * alpha - 1
    )  # Dirichlet-multinomial distribution
    for r in range(q):
        H_ng -= _log_binom(ng[r] + alpha - 1, alpha - 1)

    return H_ng


def _H_ngc_G_nc_alpha(ngc: ArrayLike, alpha: float) -> float:
    """
    Compute the entropy of a contingency table given knowledge of the column sums and the concentration parameter alpha.

    Parameters
    ----------
    ngc : list or np.ndarray
        Vector of column sums.
    alpha : float
        Concentration parameter.

    Returns
    -------
    float
        Entropy of the contingency table. (base e).
    """

    qg = int(ngc.shape[0])
    qc = int(ngc.shape[1])
    nc = np.sum(ngc, axis=0)  # Column sums

    H_ngc = 0.0
    for s in range(qc):
        H_ngc += _log_binom(
            nc[s] + float(qg) * alpha - 1, float(qg) * alpha - 1
        )  # Independent Dirichlet-multinomial distributions of the columns
        for r in range(qg):
            H_ngc -= _log_binom(ngc[r, s] + alpha - 1, alpha - 1)

    return H_ngc


def _reduced_mutual_information(contingency_table: ArrayLike) -> float:
    """
    Compute the reduced mutual information, using the Dirichlet-multinomial reduction of https://arxiv.org/pdf/2405.05393.

    Parameters
    ----------
    contingency_table : ArrayLike
        Contingency table T as a 2D NumPy array where T[r][s] counts the number of objects with label r in the ground truth g and label s in the candidate c.

    Returns
    -------
    float
        Reduced mutual information (base 2).
    """
    n: float = np.sum(contingency_table)
    nc = np.sum(contingency_table, axis=0)
    ng = np.sum(contingency_table, axis=1)

    # Range of values of the concentration parameter alpha to consider in the Dirichlet-multinomial transmissions
    min_alpha = 0.0001
    max_alpha = 10000

    # H_g
    H_qg: float = np.log(n)

    _, H_ng_G_alpha = _minimize_golden_section_log(
        lambda alpha: _H_ng_G_alpha(ng, alpha), min_alpha, max_alpha
    )  # Note that we neglect the cost to transmit the alpha parameter here, although a fixed cost would cancel in the mutual information calculation.
    H_ng_G_alpha = float(H_ng_G_alpha)
    H_g_G_ng: float = float(_log_factorial(n)) - float(np.sum(_log_factorial(ng)))
    H_g: float = H_qg + H_ng_G_alpha + H_g_G_ng

    # H_g_G_c
    _, H_ngc_G_nc_alpha = _minimize_golden_section_log(
        lambda alpha: _H_ngc_G_nc_alpha(contingency_table, alpha), min_alpha, max_alpha
    )
    H_ngc_G_nc_alpha = float(H_ngc_G_nc_alpha)
    H_g_G_c_ngc: float = float(np.sum(_log_factorial(nc))) - float(
        np.sum(_log_factorial(contingency_table.flatten()))
    )
    H_g_G_c: float = H_qg + H_ngc_G_nc_alpha + H_g_G_c_ngc

    MI: float = H_g - H_g_G_c

    return float(MI / np.log(2))  # Convert to bits (log base 2)


def normalized_mutual_information(
    input_data_1: ArrayLike | str,
    input_data_2: ArrayLike | None = None,
    *,
    variation: str = "reduced",
    normalization: str = "second",
) -> ArrayLike:
    """
    Compute the normalized mutual information between two labelings from a pair of lists, the name of a space separated file of labels,
    or a contingency table. Can specify the variation of mutual information and type of normalization.
    For the asymmetric (default) normalization, the result is reported as a fraction of the entropy of the second labeling, which is considered the ground truth.

    Raises AssertionError for invalid inputs.

    Parameters
    ----------
    input_data_1 : ArrayLike or str
        First argument. This will either be a 2D array-like which specifies the contingency table whose columns are the first labeling and rows are the second labeling,
        or a string which is the path to a file containing a list of pairs of labels,
        or a 1-D array-like of labels.
    input_data_2 : ArrayLike, optional
        Second argument. This can only be a 1-D array-like of labels in the case where the first argument is also such a list.
    variation : str, optional
        Variation of mutual information to compute. Options are:
            - "stirling": Stirling's approximation of the traditional mutual information, equal to the mutual information of the corresponding probability distributions (times the number of objects).
            - "reduced" (default): Reduced mutual information (RMI), Dirichlet-multinomial reduction of https://arxiv.org/pdf/2405.05393
            - "reduced_flat": Reduced mutual information (RMI), flat reduction of https://arxiv.org/pdf/1907.12581
            - "adjusted": Adjusted mutual information (AMI), correcting for chance: https://jmlr.csail.mit.edu/papers/volume11/vinh10a/vinh10a.pdf
            - "traditional": Traditional mutual information (MI), microcanonical
            - "stirling": Stirling's approximation of the traditional mutual information, equal to the mutual information of the corresponding probability distributions (times the number of objects).
    normalization : str, optional
        Type of normalization to apply. Options are:
            - "second" (default): Asymmetric normalization, measures how much the first labeling tells us about the second, as a fraction of all there is to know about the second labeling.
            - "first": Asymmetric normalization, measures how much the second labeling tells us about the first, as a fraction of all there is to know about the first labeling.
            - "mean": Symmetric normalization by the arithmetic mean of the two entropies.
            - "min": Normalize by the minimum of the two entropies.
            - "max": Normalize by the maximum of the two entropies.
            - "geometric": Normalize by the geometric mean of the two entropies.
            - "none": No normalization, returns the mutual information in bits.

    Returns
    -------
    float
        Normalized mutual information
    """

    # Get the contingency table
    contingency_table = _get_contingency_table(input_data_1, input_data_2)

    # Make the candidate-candidate (labeling 1) and truth-truth (label 2) contingency tables
    nc = np.sum(contingency_table, axis=0)
    ng = np.sum(contingency_table, axis=1)
    ncc = np.diag(nc)  # Candidate-candidate
    ngg = np.diag(ng)  # Truth-truth

    # Compute the mutual information between each pair of labelings
    MI_c_g, MI_g_c, MI_c_c, MI_g_g = None, None, None, None  # Values to be computed
    if variation == "stirling":
        MI_c_g = _stirling_mutual_information(
            contingency_table
        )  # Mutual information between candidate and ground truth
        MI_g_c = MI_c_g  # Symmetric measure
        MI_c_c = _stirling_mutual_information(
            ncc
        )  # Mutual information between candidate and candidate
        MI_g_g = _stirling_mutual_information(
            ngg
        )  # Mutual information between ground truth and ground truth
    elif variation == "traditional":
        MI_c_g = _traditional_mutual_information(contingency_table)
        MI_g_c = MI_c_g  # Symmetric measure
        MI_c_c = _traditional_mutual_information(ncc)
        MI_g_g = _traditional_mutual_information(ngg)
    elif variation == "adjusted":
        MI_c_g = _adjusted_mutual_information(contingency_table)
        MI_g_c = MI_c_g  # Symmetric measure
        MI_c_c = _adjusted_mutual_information(ncc)
        MI_g_g = _adjusted_mutual_information(ngg)
    elif variation == "reduced":
        MI_c_g = _reduced_mutual_information(contingency_table)
        MI_g_c = _reduced_mutual_information(
            contingency_table.T
        )  # Transpose to get the ground truth as the first labeling
        MI_c_c = _reduced_mutual_information(ncc)
        MI_g_g = _reduced_mutual_information(ngg)
    elif variation == "reduced_flat":
        MI_c_g = _reduced_flat_mutual_information(contingency_table)
        MI_g_c = _reduced_flat_mutual_information(contingency_table.T)
        MI_c_c = _reduced_flat_mutual_information(ncc)
        MI_g_g = _reduced_flat_mutual_information(ngg)
    else:
        raise ValueError(f"Unknown variation type: {variation}")

    # Compute the normalized mutual information
    if normalization == "second":
        # Asymmetric normalization, measures how much the first labeling tells us about the second, as a fraction of all there is to know about the second labeling
        return MI_c_g / MI_g_g if MI_g_g > 0 else 0
    if normalization == "first":
        # Asymmetric normalization, measures how much the second labeling tells us about the first, as a fraction of all there is to know about the first labeling
        return MI_g_c / MI_c_c if MI_c_c > 0 else 0
    if (
        normalization == "mean"
    ):  # Note that the numerators of these symmetric measures are non-standard in order to account for asymmetries in the calculated MI_c_g vs MI_g_c
        return (MI_c_g + MI_g_c) / (MI_c_c + MI_g_g) if (MI_c_c + MI_g_g) > 0 else 0
    if normalization == "min":
        return (
            min(MI_c_g, MI_g_c) / min(MI_c_c, MI_g_g) if min(MI_c_c, MI_g_g) > 0 else 0
        )
    if normalization == "max":
        return (
            max(MI_c_g, MI_g_c) / max(MI_c_c, MI_g_g) if max(MI_c_c, MI_g_g) > 0 else 0
        )
    if normalization == "geometric":
        return (
            np.sqrt(MI_c_g * MI_g_c) / np.sqrt(MI_c_c * MI_g_g)
            if (MI_c_c * MI_g_g) > 0
            else 0
        )
    if normalization == "none":
        return MI_c_g  # Return the mutual information in bits without normalization (note that this may not be symmetric for the reduced measures)
    raise ValueError(f"Unknown normalization type: {normalization}")


def mutual_information(
    input_data_1: ArrayLike | str,
    input_data_2: ArrayLike | None = None,
    *,
    variation: str = "reduced",
) -> ArrayLike:
    """
    Compute the mutual information between two labelings from a pair of lists, the name of a space separated file of labels,
    or a contingency table. Can specify the variation of mutual information to compute.

    Raises AssertionError for invalid inputs.

    Parameters
    ----------
    input_data_1 : ArrayLike or str
        First argument. This will either be a 2D array-like which specifies the contingency table whose columns are the first labeling and rows are the second labeling,
        or a string which is the path to a file containing a list of pairs of labels,
        or a 1-D array-like of labels.
    input_data_2 : ArrayLike, optional
        Second argument. This can only be a 1-D array-like of labels in the case where the first argument is also such a list.
    variation : str, optional
        Variation of mutual information to compute. Options are:
            - "reduced" (default): Reduced mutual information (RMI), reduction of https://arxiv.org/pdf/2405.05393, note that this can be (slightly) asymmetric.
            - "reduced_flat": Reduced mutual information (RMI), flat reduction of https://arxiv.org/pdf/1907.12581
            - "adjusted": Adjusted mutual information (AMI), correcting for chance: https://jmlr.csail.mit.edu/papers/volume11/vinh10a/vinh10a.pdf
            - "traditional": Traditional mutual information (MI), microcanonical
            - "stirling": Stirling's approximation of the traditional mutual information, equal to the mutual information of the corresponding probability distributions (times the number of objects).

    Returns
    -------
    float
        Mutual information value in bits (base 2).
    """

    return normalized_mutual_information(
        input_data_1,
        input_data_2,
        variation=variation,
        normalization="none",
    )



"""
Copyright (c) 2025 Max Jerdee. All rights reserved.

clustering-mi: Compute the mutual information between two clusterings of the same objects
"""

from __future__ import annotations

# first party imports
from clustering_mi._input_output import _get_contingency_table
from clustering_mi.mutual_information import (
    mutual_information,
    normalized_mutual_information,
)

from ._version import version as __version__

__all__ = [
    "__version__",
    "_get_contingency_table",
    "mutual_information",
    "normalized_mutual_information",
]
